/* iobench.c
/* Copyright (c) 1988, Prime Computer, Inc., Natick, MA 01760
/*                     All Rights Reserved
*/

/* Note: if you get a compilation error because include file <utsname.h>
   can't be found, it means the uname() system call isnt' supported on
   your system.  In this case, add "-DNOUNAME" to EXTRA_CFLAGS in your
   Makefile wrapper.

   This is driver program for IOBENCH test.  It creates and populates the
   necessary files, launches the daemon processes that do the benchmark
   proper, and then collects, summarizes, and reports the benchmark
   results.  Although IOBENCH is meant to be operating system independent,
   it does things in the UNIX style.  Unless absolutely necessary, non
   portable UNIX features are not used.

   The overall plan for the benchmark is to create NF files each containing
   NR RECLEN-byte records.  The records have implicit keys, which are longs
   in the range 0...NR-1.  The only interesting part of each record is a
   long counter.  NU daemon users are created.  Each user does
   "transactions" consisting of R reads and W writes with an optional CP
   load.  IOBENCH can be run in "timed" mode, in which all of the users do
   as many transactions as they can in the specified time interval, or in
   "counted" mode, in which case each user does its share (1/NU) of the
   specified number of transactions as quickly as it can.  In timed mode,
   we're interested in how many transactions were done; in counted mode,
   we're interested in how long each user took.  See "iouser.c" for details
   on transactions.

   Each of the user daemons writes a result record to record K of the
   results file containing statistics such as number, of transactions
   completed, CP time used, etc.  Detailed statistics of system activity
   are gathered by phantom NU+1, which runs PRIMOS usage, UNIX sar, etc.
   and writes its results into a file.

   Each of the user processes "locks" the records it reads and writes
   during the period in which it is using them.  This locking is a
   simulation of what would be done in a "real" data management system; the
   only requirement is that no other user be able to access the locked
   record until locking user releases the lock.  IOBENCH is distributed
   with locking code implemented for some architectures in terms of atomic
   instructions such as "test and set"; this code is activated by compiling
   IOBENCH with REALLOCKING defined.  If you're porting IOBENCH to a new
   architecture, you'll have to implement about half a page of assembly
   code.  You can build IOBENCH with REALLOCKING not defined, in which case
   "locking" is implemented via ordinary non-atomic C code; this
   approximates the overhead that would be incurred from real locking.  See
   "iolocks.c" for more information on locking.

   IOBENCH is controlled by an input file that contains the following
   information (see "ioparams.c" for more details):

         runid
         number of files
         number of records (or total file length)
         record size
         number of users
         locks per file
         lock sleep time
         start delay
         run duration
         end delay
         cpu work parameter
         reads per transaction
         writes per transaction (must be <= reads per transaction)
         pathname for file 1      OR   @filespath
         pathname for file 2
         ...

   The file pathnames are provided in the control file so that IO load can
   be spread across multiple disks and controllers.  If "@filespath" is
   specified, file names are read from file with specified pathname instead
   of from original control file.

   Synchronization and control of the user processes is done by an ad hoc
   scheme that uses the file system and avoids operating system
   dependencies.

   Startup of the background processes is controlled by a sync file.  Each
   background task does its initialization and then waits until the
   controlling process creates the sync file.  As soon as the sync file
   appears, the background processes start their timed period.

   As soon as each background process starts running, it creates a unique
   run file using the name passed to it by the controlling process.  The
   background processes delete their unique run files as the last thing
   they do.  Thus, the existance of any run file means the process
   associated with that task is still active and the date in the results
   file is not yet valid.

   IOBENCH is a "relative", not an "absolute" benchmark.  The most useful
   way to use IOBENCH is to define a spectrum of runs that vary key
   parameters such as number of users, CP work component, or record length
   over a range of 10-20 values that cover both IO limited and CP limited
   performance.  Do the set of runs on a base configuration, and then redo
   them on a different configuration (e.g.  different hardware platform,
   different OS, different IO controller, etc) and compare the results
   against the base configuration.

   The name of the input control file is used to generate the names of two
   output files.  control.res contains the results of the run and
   control.use contains system usage data for the run.  An optional summary
   file can be created containing one line for each test configuration.
   If the -l flag is specified, iobench creates a log file (suffix ".log")
   that contains a series of lines indicating successful completion of
   major stages in the benchmark run.  Unlike the terminal output or the
   results file, the log will for a correct run will always contain the
   same contents and can be compared against the log file of another
   correct run.

   The -c option causes iobench to show a series of checkpoints.  These
   should help you determine if performance was uniform or if there were
   hot spots.

   Initial Version: 10 August 1988 by Barry Wolman

          Modified: 16 August 1988 by Barry Wolman
              1. extend file lengths
              2. run iostat background process
              3. query environment

          Modified: 20 August 1988 by Barry Wolman
              1. Backfit results of port to UNIX
              2. Add version ID to output

          Modified: 28 August 1988 by Barry Wolman
              1. Deal with case where a background process doesn't terminate
              2. Speed up termination check
              3. Store number of users in results file

          Modified: 29 August 1988 by Barry Wolman
              1. Print process ids of background tasks
              2. Kill processes if error
              3. Add work parameter
              4. Fixed bug in file population (needed seek to EOF) when making
                 existing file larger

          Modified: 6 September 1988 by Barry Wolman
              1. Print volume of data processed
              2. Identify run file in release message

          Modified: 10 September 1988 by Barry Wolman
              1. Close output file before spawning UNIX daemons
              2. Reopen output file after UNIX daemons finish
              3. Summarize file closing and opening due to file unit limits
              4. Show run duration instead of finish time
              5. Add debug flag to control printing of process ids
              6. Use SIGTERM instead of SIGKILL for terminating user procs
              7. Catch SIGTERM & SIGINT (UNIX) and QUIT$ PRIMOS and do a
                 cleanup

          Modified: 16 September 1988 by Barry Wolman
              1. Calculate adjusted totals if some runs didn't finish
              2. Kill phantoms if users quits (PRIMOS)
              3. Sleep 10 seconds and LOOP 10 times when checkin termination
              4. Print run summary in grep-able form

          Modified: 26 September 1988 by Barry Wolman
              1. Make PRIMOS Phantom logout messages come out AFTER all of
                 our output.
              2. Eliminate phantom logout message if not debug (PRIMOS)

          Modified: 27 September 1988 by Barry Wolman
              1. Set RW lock on results file (PRIMOS)
              2. Print statistics gathered by phantoms

          Modified:  2 October 1988 by Barry Wolman
              1. Skip IO if calibrate flag set
              2. Sleep 30 seconds when checking termination
              3. Rewrote termination check to sleep AFTER check not BEFORE

          Modified: 15 February 1989 by Tom Olson
              1. Added transaction grouping,
              2. Added read/write ratio (n_reads and n_writes) parameters
              3. Compute transaction response time distribution statistics.

          Modified: 27 February 1989 by Barry Wolman
              1. Call init_params to initialize params from defaults and
                 environment
              2. Parse keyword style contro lfile
              3. Create interim control file for user processes
              4. Write line to summary file
              5. Move in check_params and support subroutines from iosubs
              6. Minor cleanup

         Modified: 28 March 1989 by Barry Wolman
              1. Correct calculation of total MB (Olson's Bug!)
              2. Show TPS summary even if -t not specified, -t now controls
                 only display of detailed histogram
              3. Allow input file to be empty

         Modified:  3 April 1989 by Barry Wolman
              1. Make input file optional
              2. Add -a, -h, and -u options
              3. Use UPPERCASE for temp file names
              4. Delete RESULTS file at end of run now that prt_results
                 is defunct.

         Modified: 25 May 1989 by Barry Wolman
              1. Add variant capability

         Modified: 24 August 1989 by Barry Wolman
              1. Don't start server if variant is 0
              2. Don't compile server code if no message queue facility

         Modified:  2 May 1990 by Barry Wolman
              1. Keep going with reduced number if users if all user
                 processes can't be started

         Modified: 18 May 1990 by Barry Wolman
              1. LOCKS=0 omits locking
              2. Pass delay parameter to usage process
              3. Show number of lock hits

         Modified:  6 June 1990 by Barry Wolman
              1. Adapt to SPEC Makefile conventions

         Modified: 20 June 1990 by Barry Wolman
              1. Keep user & sys times separate, report total times
              2. Move parameter processing to ioparams.c
              3. Use filelen instead of total_bytes
              4. All durations in seconds
              5. Call ioflush() to force out cached pages
              6. Loop over output files

         Modified: 16 July 1990 by Barry Wolman
              1. Add mounting
              2. Add logging

         Modified: 24 July 1990 by Barry Wolman
              1. Rewrite error messages to use ERR mechanism
              2. Change gripe calls to ERR macro
              3. Unmount file systems before starting if mounts non-zero
              4. Fix bug setting pid[i]
              5. Use mode change to 0666 as indication that iouser has finished
              6. Read log from runfile, use for validation
              7. Add -f (full validation) & -k (keep files) options

         Modified: 30 July 1990 by Barry Wolman
              1. Add ulimit check & set
              2. Put command line in output files
              3. Add setcleanup call & moved up signal calls
              4. Add remount option

         Modified: 6 August 1990 by Barry Wolman
              1. Fixed help message slightly for -l
              2. Add some error checks on system calls

         Modified: 10 August 1990 by Barry Wolman
              1. Fix definition of cleanup() & malloc() to satify ANSI
                 compilers

         Modified: 16 August 1990 by Barry Wolman
              1. Partitioned main
              2. Minor changes to run on PRIMOS
              3. Call to validate was overwriting err flag setting
              4. Add alternate wait_for_users_to_finish that uses
                 wait system call

         Modified: 21 August 1990 by Barry Wolman
              1. Minor edits
              2. Add max_time to transaction statistics (required reordering
                 output sections)
              3. Add warning if GTOD undefined
              4. Add -brief option so warning can be suppressed
              5. Move deletion of IORUN files to delete_run_files routine,
                 turn on keep flag if validation errors detected
              6. Move formation of IORUNn into runfile()
              7. Don't validate if write=0
              8. Use uname to annotate run

         Modified: 2 September 1990 by Barry Wolman
              1. Show breakdown of user & system times
              2. Show TPS/user

         Modified: 8 October 1990 by Barry Wolman
              1. Look for paths to mount & umount in environment

         Modified: 15 October 1990 by Barry Wolman
              1. Print additional info when ulimit can't be changed
              2. Removed exit(0) from cleanup(), put it at end of main()
              3. Print message suggesting user try again if he gets
                 validation errors

         Modified: 19 October 1990 by Barry Wolman
              1. Print debug info during validation if debug flag set

         Modified: 25 October 1990 by Barry Wolman
              1. Use SPEC standard mount_disks and unmount_disks
              2. Use check argument in dounmounts()
              3. Don't use the word "transaction" in SPEC version output
              4. Delete the large files if we run out of space
              5. Use getdevice() to check validity of specified devices

         Modified: 7 November 1990 by Barry Wolman
              1. Read seeds file, pass starting random seed to iouser
              2. Kill all iouser processes if we can't create them all
              3. Fix bug in lseek to extend file length
              4. Check for non-normal iouser termination
              5. Few POSIX cleanups
              6. Fix bug in call to statlocks

         Modified: 3 December 1990 by Barry Wolman
              1. Finally fixed the 0 flush time problem!

*/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "ioincl.h"
#include "iodefaults.h"
#include "ioserver.h"

#if UNIX
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#ifndef NOUNAME
#include <sys/utsname.h>
#endif
void cleanup();
#define APPEND_MODE     "a"
#define CREAT_MODE      0666
#endif

#if PRIMOS
#define SIGTERM 15
fortran void mkon$p(),lon$cn(),lon$r();
fortran void ignore_logouts();
fortran void cleanup();
#define APPEND_MODE     "wa"
#define CREAT_MODE      3
#define O_RDONLY        0
#define O_RDWR          2
#define SEEK_SET        0
#define SEEK_END        2
#endif

#ifdef SPEC
#define TRANS      "cycle"
#define CAPTRANS   "Cycle"
#define NUMTRANS   "#Cycles"
#define NUMTRAN    "#CYCS"
#define TPS        "CPS"
#else
#define TRANS      "transaction"
#define CAPTRANS   "Transaction"
#define NUMTRANS   " #TRANS"
#define NUMTRAN    "#TRAN"
#define TPS        "TPS"
#endif

#define SYNCFILE   "IOSYNCFILE"
#define RUNFILE    "IORUN"
#define RESULTS    "IORESULTS"
#define CTLFILE    "IOCONTROLFILE"
#define OUTFILE    "iobench"

#define MAXLOOPS   10

/* Following macro is used to iterate over the set of output files */

#define OFLOOP for(z=0; of[z]; z++)
#define LOG(x) if (genlog) fprintf(logfile,"%s\n",(x))

int unit[MAX_FILES];

struct control control;

struct logrec *userlog[MAX_USERS];
long userlogsize[MAX_USERS];

long start_time,finish_time;
long *seeds,nseeds;

FILE *of[4], *logfile, *fopen();

extern char *version,*flags,*compdate;

char *getenv(),*copystr(),*runfile(),*getcwd(),*getdevice();
void *malloc(),*calloc();
long ioflush(),*read_seeds();

int debug = 0, show_tps = 0, append = 0, help = 0, genlog = 0, usage = 0;
int show_checkpoints = 0, full = 0, keep = 0,  remount = 0, brief = 0;
int old_ulimit = 0, new_ulimit = 0, inited_locks = 0;
int had_validation_error = 0;

char *iosummary = NULL;
char *iooutput = NULL;
char *lnspec = NULL;
char logname[256],outname[256],usename[256];

char *mountpath = "./mount_disks";
char *umountpath = "./umount_disks";
char *ioseeds = "ioseeds";

int pid[MAX_USERS],finished[MAX_USERS];
int spawned_users = 0;
int smsgid,rmsgid,mypid,serverpid,user;
#if UNIX
struct msgstruct smsg,rmsg;
#endif
char *instrs[] = {
    "Usage: iomain [options] [control_file]",
    "",
    "Options are: -a         append to output file",
    "             -b         brief mode, omit some warnings",
    "             -c         show checkpoints",
    "             -d         set debug flag",
    "             -f         do full validation (doubles run time)",
    "             -h         help, print this message",
    "             -k         keep intermediate files",
    "             -l[file]   generate logfile, file is optional",
    "             -ofile     use 'file' as prefix of output files",
    "             -r         remount file systems at end of run",
    "             -sfile     write summary line to 'file'",
    "             -t         show TPS statistics",
    "             -u         collect system usage statistics",
#ifdef MESSSAGEQUEUES
    "             -vN        set variant to N, where N is octal #",
#endif
    "",
    "Control file can be omitted if all parameters are specified via",
    "environment or defaults.",
    "",
    "Contact Barry Wolman (barry@s66.prime.com) if you have questions.\n",
    NULL};

/* Following variables are used by routines that summarize and output
   the run data */

int errs,hits,nc,no,ntrans,total_nc,total_no,failed,opens[MAX_FILES],
 closes[MAX_FILES],uses[MAX_FILES],worst_rec[MAX_FILES],worst_user[MAX_FILES];

long et,maxet,ft,totaltime,cum_resp_dist[26],reads[MAX_FILES],
 writes[MAX_FILES];

double sum_time[MAX_FILES],sumsq_time[MAX_FILES],worst_time[MAX_FILES],
 usert1,usert2,syst1,syst2,mb,avg_time,stddev_time,avg_resp,server_usert,
 server_syst,flush_usert,flush_syst,total_avg_resp,tps,totalcp;

struct checkpoint checkpoint[NUMBER_OF_CHECKPOINTS+1];

main(argc,argv)
char **argv;
    {
    register int i;
    int rfid,z,save_argc;
    char *p;
    char **save_argv,wd[256];

#if UNIX
#ifndef NOUNAME
    struct utsname uinfo;
#endif
#endif

    /* Ensure proper cleanup.  Note: PRIMOS condition handlers (e.g. for
       QUIT$), are associated with a specific stack frame and automatically
       are revoked when the stack frame is destroyed.  Thus, all signal
       or condition related code should be left HERE and SHOULD NOT be
       moved into a separate procedure */

    setcleanup(cleanup);

#if UNIX
    signal(SIGTERM,cleanup);
    signal(SIGINT,cleanup);
#endif

#if PRIMOS
    mkon$p("QUIT$",5,cleanup);

    if (debug) lon$cn((short)0);        /* Turn off notification */
    else mkon$p("PH_LOGO$",8,ignore_logouts);
#endif

    /* End of code region that shouldn't be moved */

    /* Check for SUMMARY and OUTPUT specified in environment so that
       specification on command line can override these */

    if (p = getenv("SUMMARY")) iosummary = copystr(p);
    if (p = getenv("OUTPUT")) iooutput = copystr(p);

    /* Check for override of default mount & umount paths.  Environment
       variable must be defined and point to a non-zero length string */

    if ((p = getenv("MOUNT")) && *p) mountpath = copystr(p);
    if ((p = getenv("UMOUNT")) && *p) umountpath = copystr(p);

    /* Check for override of ioseeds file pathname */

    if ((p = getenv("IOSEEDS")) && *p) ioseeds = copystr(p);

    /* Now process arguments and options */

    errs = 0;

    control.variant = 0;
    save_argc = argc;
    save_argv = argv;

    errs = process_options(&argc,&argv);

    if (help || errs) exit(1);

    read_seeds(ioseeds);

    open_output_files(argc,argv);

    time(&start_time);

    LOG("Start");

    init_params();
    if (argc > 1 && parse_control(argv[1])) exit(1);

    if (check_params()) exit(1);

    if (control.n_users > nseeds) {
         errno = 0;
         MSGSS("You requested %d users, but there are only %d random number seeds",
          control.n_users,nseeds);
         ERRSS("Rerun with no more than %d users, or use genseeds utility to generate\n\
         a new ioseeds file with at least %d values and then return",nseeds,control.n_users);
         }

    /* Either LOCKS=0 (new style) or SLEEP<0 (old style) surpresses locking */

    if (control.lock_sleep_time < 0) control.locks_per_file = 0;

    /* We can't assume that user processes will have access to environment
       variables, so we'll create a complete copy of control file */

    if (dump_control(CTLFILE)) ERR("Could not copy control file");

    OFLOOP print_control(of[z],save_argc,save_argv);

#if UNIX
#ifndef NOUNAME
    uname(&uinfo);

    OFLOOP fprintf(of[z],"Machine %s, Nodename %s, Sysname %s, Release %s, Version %s\n",
            uinfo.machine, uinfo.nodename, uinfo.sysname, uinfo.release,
            uinfo.version);
#endif
#endif

    /* Look for optional environment variables that define system
       configuration */

    OFLOOP {
         if (p = getenv("SYSTEM")) fprintf(of[z],"System is %s\n",p);
         if (p = getenv("MEMORY")) fprintf(of[z],"Memory is %s\n",p);
         if (p = getenv("CONTROLLERS")) fprintf(of[z],"Controllers is %s\n",p);
         if (p = getenv("BUFFERS")) fprintf(of[z],"Buffers is %s\n",p);
         if (p = getenv("COMMENT")) fprintf(of[z],"Comment is %s\n",p);
         }

#if UNIX
    adjust_ulimit();
#endif

    LOG("Got Parameters");

    if (control.mounts) {

         if (p = getdevice(getcwd(wd,256))) {

              /* Make sure we haven't been asked to unmount the current
                 working directory */

              for(i=0;i<control.mounts;i++)
                   if (strcmp(p,control.device[i]) == 0) {
                        MSGS("Sorry, I can't unmount device %s",control.device[i]);
                        MSGS("because the current working directory %s",wd);
                        ERR("is mounted on it!  Specify another device and re-run.");
                        }
              }


         /* Unmount specified devices */

         dounmounts(1);

         /* Mount specified dirs on devices */

         domounts();
         }

    LOG("Mounted");

    initialize_files();

    printf("Finished file population\n");
    LOG("Populated");

    /* Users = 0 means just file population */

    if (control.n_users == 0) return;

#ifdef MESSAGEQUEUES
    if (control.variant) start_server();
    else serverpid = 0;
#endif

    /* Initialize the lock manager */

    initlocks(control.n_users,control.n_files,control.locks_per_file,
     control.variant & LOCKING_VARIANT);
    inited_locks = 1;

    /* Delete the sync file used to control start up of users */

    unlink(SYNCFILE);

    /* Create the results file */

    create_results_file();

    LOG("Initialized Results");

#if UNIX
    if (fclose(of[1])) ERRS("Could not close %s",outname);

    of[1] = NULL;
#endif

    if (start_user_processes()) ERR("Could not create all iouser processes.");

    if (usage) {

         /* Start background process to do system monitoring passing it
            arguments specifying
             1) number of seconds sleep before starting
             2) number of seconds to gather usage statistics
             3) name of output file to be written
             4) append flag */

         monitor_system(control.start_delay,control.duration,usename,append);
         spawned_users++;
         }

    LOG("Spawned Users");

    /* All the users are now running.  They will do their initialization
       and then wait for us to create the sync file before proceeding
       with benchmark proper. */

    sleep(control.start_delay);

    if (creat(SYNCFILE,CREAT_MODE) < 0) {
#if PRIMOS
         for (i = 1; i <= control.n_users; i++) killuser(i,pid);
#endif
         ERRS("Could not create %s",SYNCFILE);
         }

    printf("Released background processes for %s at %s\n",iooutput,
     cvtime(time(0),0));

    /* User processes should now be running the benchmark proper.  Wait
       for them to finish */

    wait_for_users_to_finish();

    unlink(SYNCFILE);

#if UNIX
    if ((of[1] = fopen(outname,"a")) == NULL) ERRS("Could not reopen %s",outname);
#endif

    if (errs == 0) LOG("Users Finished");

    hits =  statlocks(of[1]);

    termlocks();
    inited_locks = 0;

#ifdef MESSAGEQUEUES
    if (serverpid) stop_server();
    else server_usert = server_syst = 0.0;
#else
    server_usert = server_syst = 0.0;
#endif

    /* Flush any cached pages that need to be written to disk and
       remember how long that took */

    errs = 0;
    if ((ft = ioflush(&flush_usert,&flush_syst)) < 0) errs++;

    LOG("Flushed");

    /* Reopen the benchmark files so we can validate the run */

    if (control.mounts) domounts();

    for(i = 0; i < control.n_files; i++)
         if ((unit[i] = open(control.path[i],O_RDONLY)) < 0)
              ERRS("Could not reopen %s",control.path[i]);

    if ((rfid = open(RESULTS,O_RDONLY)) < 0) ERRS("Could not reopen %s",RESULTS);

    output_user_statistics(rfid);

    if (validate()) errs++;

    for(i = 0; i < control.n_files; i++) close(unit[i]);

    if (control.mounts  && !remount) dounmounts(0);

    if (show_checkpoints) output_checkpoints();

    /* total benchmark time is sum of max actual time and flush time */

    totaltime = maxet + ft;

    /* Note - small test cases can yield runs that execute in 0 seconds ! */

    if (totaltime == 0) totaltime = 1;

    if (total_nc + total_no) {
         OFLOOP fprintf(of[z],"\n%d file opens and %d file closes\n",total_no,total_nc);
         }

    output_file_statistics();

    time(&finish_time);

    output_totals();

    output_transaction_statistics(rfid);

    if (failed) adjust_results();

    write_summary_line();

    if (failed+errs == 0) LOG("Validated Results");

#if PRIMOS
    if (!debug) {

         /* Wait for all phantom logout signals to arrive */

         for(i=0;i < 60 && spawned_users > 0;i++) sleep(1);
         }
#endif

    delete_run_files();

    if (had_validation_error) {
         fprintf(stderr,"There were %d validation errors in this test.\n\
Since IOBENCH was built with pseudo-locking, code that should be a critical\n\
region isn't.  Chances are if you rerun your test, the validation error(s)\n\
won't reoccur.  Before redoing the test, be sure to save the files 'IORUN*'\n\
produced by the current test so they can be examined at a later time.  If\n\
you're doing a SPEC Makefile run, just rename the 'result' directory to keep\n\
it from being removed by the next test;  otherwise, copy 'IORUN*' and the\n\
printable files produced by the current run to a new, empty directory. If you\n\
continue to get validation errors when you rerun your test, refer to the\n\
documentation in 'iobench.out' for information on how to use the 'prlog'\n\
tool to analyze the saved IORUN* files.\n",had_validation_error);
         }

    OFLOOP fprintf(of[z],"Finished %s at %s\n",iooutput,cvtime(finish_time,0));
    fclose(of[1]);                      /* .. */
    if (genlog) fclose(logfile);        /* .. */

#if PRIMOS
    if (debug) lon$cn((short)1);        /* Turn on notification */
#endif

    exit(had_validation_error ? 1 : 0);
    }

/* Process options in command line, return number of errors found.
   Arguments are POINTERS to argc & argv that were passed to main */

process_options(argcp,argvp)
int *argcp;
char ***argvp;
    {
    register int errs = 0,i;

    while(*argcp > 1 && (*argvp)[1][0] == '-') {
         switch((*argvp)[1][1]) {

         case 'a':
         case 'A':
                   append++;
                   break;

         case 'b':
         case 'B':
                   brief++;
                   break;

         case 'c':
         case 'C':
                   show_checkpoints++;
                   break;

         case 'd':
         case 'D':
                   debug++;
                   break;

         case 'f':
         case 'F':
                   full++;
                   break;

         case 'h':
         case 'H':
                   for(i = 0; instrs[i]; i++) fprintf(stderr,"%s\n",instrs[i]);
                   help++;
                   break;

         case 'k':
         case 'K':
                   keep++;
                   break;

         case 'l':
         case 'L':
                   genlog++;
                   if ((*argvp)[1][2]) lnspec = (*argvp)[1]+2;
                   break;

         case 'o':
         case 'O':
                   iooutput = (*argvp)[1][2] ? (*argvp)[1]+2 : NULL;
                   break;

         case 'r':
         case 'R':
                   remount++;
                   break;

         case 's':
         case 'S':
                   iosummary = (*argvp)[1][2] ? (*argvp)[1]+2 : NULL;
                   break;

         case 't':
         case 'T':
                   show_tps++;
                   break;

         case 'u':
         case 'U':
                   usage++;
                   break;

#ifdef MESSAGEQUEUES
         case 'v':
         case 'V':
                   if (sscanf((*argvp)[1]+2,"%o",&control.variant) != 1) {
                        fprintf(stderr,"iobench: '%s' is not a valid variant\n",
                         (*argvp)[1]+2);
                        errs++;
                        }

                   break;
#endif

         default:  fprintf(stderr,"iobench: '%c' is not a valid option\n",(*argvp)[1][1]);
                   errs++;
                   }

         (*argvp)++;
         (*argcp)--;
         }

    return(errs);
    }

/* Read ioseeds file */

long *read_seeds(name)
char *name;
    {
    register FILE *sp;
    register int i;

    if ((sp = fopen(name,"r")) == NULL) ERRS("Could not open %s",name);

    if (fscanf(sp,"%d\n",&nseeds) != 1) ERR("Could not determine number of seeds");

    if ((seeds = (long *)calloc(nseeds,sizeof(long))) == NULL)
         ERRS("Could not allocate space for %d seeds",nseeds);

    for (i= 0; i < nseeds; i++)
         if (fscanf(sp,"%d\n",&seeds[i]) != 1) ERRS("Could not read seed %d",i);

    fclose(sp);
    return(seeds);
    }

/* Form names of output files and open them */

open_output_files(argc,argv)
int argc;
char **argv;
    {

    /* If no output file specified via environment, use name of input file
       if present, otherwise use default */

    if (iooutput == NULL) iooutput = (argc > 1 ? argv[1] : OUTFILE);

    sprintf(outname,"%s.res",iooutput);
    sprintf(usename,"%s.use",iooutput);

    sprintf(logname,"%s.log",lnspec ? lnspec : iooutput);

    if (append) {
         of[1] = fopen(outname,APPEND_MODE);
         }
    else of[1] = fopen(outname,"w");

    if (of[1] == NULL) ERRS("Could not (re)open %s",outname);

    of[0] = stdout;
    of[2] = of[3] = NULL;

    if (genlog)
         if ((logfile = fopen(logname,"w")) == NULL)
              ERRS("Could not create logfile %s",logname);
    }

/* Populate a file */

populate(fn,rp)
register int fn;
register struct rec *rp;
    {
    register int curr_size,left;
    int z;
    long t1,t2;
    double usert1,usert2,syst1,syst2;

    if ((curr_size = fsize(control.path[fn])) >= control.filelen)
         return(close(unit[fn]));

    cptimes(&usert1,&syst1);
    time(&t1);

    if (curr_size) {
         OFLOOP fprintf(of[z],"Increasing size of %s from %d to %d bytes\n",control.path[fn],
          curr_size,control.filelen);

         if (lseek(unit[fn],0,SEEK_END) < 0)                /* Seek to EOF */
              ERRS("Could not lseek to EOF on %s",control.path[fn]);
         }
    else {
         OFLOOP fprintf(of[z],"Populating %s to %d bytes\n",control.path[fn],
          control.filelen);
         }

    while(curr_size+INITRECLEN <= control.filelen) {
         if (write(unit[fn],rp,INITRECLEN) < 0) {
              MSGSS("Write failed at byte %d of %s",curr_size,control.path[fn]);
              return(-1);
              }

         curr_size += INITRECLEN;
         }

    if (left = control.filelen - curr_size) {
         if (write(unit[fn],rp,left) < 0) {
              MSGSS("Write failed at byte %d of %s",left,control.path[fn]);
              return(-1);
              }
         }

    cptimes(&usert2,&syst2);
    time(&t2);
    OFLOOP fprintf(of[z],"%d seconds, %.3f user CP seconds, %.3f system CP seconds\n",t2-t1,usert2-usert1,syst2-syst1);

    return(close(unit[fn]));
    }

/* Print contents of control file */

print_control(fid,argc,argv)
register FILE *fid;
int argc;
char **argv;
    {
    register int i;
    fprintf(fid,"Started %s at %s\n",iooutput,cvtime(start_time,0));
    fprintf(fid,"IOBENCH version %s built on %s\n",version,compdate);
    fprintf(fid,"IOBENCH build flags: %s\n",flags);
    fprintf(fid,"Logging to file %s\n",logname);
    if (iosummary) fprintf(fid,"Summary will be written to file %s\n",iosummary);
    fprintf(fid,"Command line is '%s",argv[0]);
    for(i=1;i<argc;i++) fprintf(fid," %s",argv[i]);
    fprintf(fid,"'\n");
    fprintf(fid,"Run id is %d\n",control.runid);
    if (control.variant) fprintf(fid,"Variant is %o\n",control.variant);
    if (control.runtype == TIMED_RUN) fprintf(fid,"Timed run for %d seconds\n",control.duration);
    else fprintf(fid,"Counted run for %d %ss\nEstimated duration = %d seconds\n",
          control.trans,TRANS,control.duration);
    fprintf(fid,"Number of files = %d\n",control.n_files);
    fprintf(fid,"Number of records = %d\n",control.n_recs);
    fprintf(fid,"Record length = %d\n",control.reclen);
    fprintf(fid,"File length = %d\n",control.filelen);
    fprintf(fid,"Number of users = %d\n",control.n_users);
    fprintf(fid,"Locks per file = %d\n",control.locks_per_file);
    fprintf(fid,"Lock sleep time = %d\n",control.lock_sleep_time);
    fprintf(fid,"Start delay = %d\n",control.start_delay);
    fprintf(fid,"End delay = %d\n",control.end_delay);
    fprintf(fid,"Work parameter = %d\n",control.param);
    fprintf(fid,"Reads = %d\n",control.n_reads);
    fprintf(fid,"Writes = %d\n",control.n_writes);

    if (control.calibrate)
         fprintf(fid,"Calibration run, no IO will be done\n");

    if (control.locks_per_file == 0)
         fprintf(fid,"No locking will be done\n");

    for(i=0; i < control.n_files; i++)
         fprintf(fid,"path[%d] = %s\n",i,control.path[i]);

    if (control.mounts) {
        fprintf(fid,"Number of file systems to mount = %d\n",control.mounts);
        for(i=0; i < control.mounts; i++)
             fprintf(fid,"device[%d] = %s, dir[%d] = %s\n",i,control.device[i],
              i,control.dir[i]);
        }
    }

/* Dump copy of control information into specified file.  Return 1 if error,
   0 otherwise */

dump_control(fname)
char *fname;
    {
    register FILE *fid;
    register int i;

    unlink(fname);

    if ((fid = fopen(fname,"w")) == NULL) return(1);

    fprintf(fid,"%d\n",control.runid);
    fprintf(fid,"%d\n",control.runtype);
    fprintf(fid,"%d\n",control.n_files);
    fprintf(fid,"%d\n",control.n_recs);
    fprintf(fid,"%d\n",control.reclen);
    fprintf(fid,"%d\n",control.filelen);
    fprintf(fid,"%d\n",control.n_users);
    fprintf(fid,"%d\n",control.locks_per_file);
    fprintf(fid,"%d\n",control.lock_sleep_time);
    fprintf(fid,"%d\n",control.start_delay);
    fprintf(fid,"%d\n",control.duration);
    fprintf(fid,"%d\n",control.end_delay);
    fprintf(fid,"%d\n",control.param);
    fprintf(fid,"%d\n",control.n_reads);
    fprintf(fid,"%d\n",control.n_writes);
    fprintf(fid,"%d\n",control.trans);
    fprintf(fid,"%d\n",control.variant);

    for(i=0; i < control.n_files; i++)
         fprintf(fid,"%s\n",control.path[i]);

    return(fclose(fid) < 0 ? 1 : 0);
    }

/* Adjust ulimit to accomodate working files */

#if UNIX
adjust_ulimit()
    {
    register int z;

    new_ulimit = old_ulimit = ulimit(1,0);
    if (control.filelen > old_ulimit * 512) {
         new_ulimit = control.filelen / 512;
         if (control.filelen % 512) new_ulimit++;

         if (ulimit(2,new_ulimit) < 0) {
              MSGSS("Could not change ulimit from %d to %d",old_ulimit,new_ulimit);
              errno = 0;
              ERRS("Rebuild UNIX with larger ulimit and rerun test,\n\
         or rerun test logged in as root,\n\
         or rerun test using a record length <= %d",old_ulimit*512);
              }

         OFLOOP {
              fprintf(of[z],"Increased ulimit from %d to %d\n",old_ulimit,
               new_ulimit);
              }
         }
    }
#endif

#ifdef MESSAGEQUEUES

/* Start server process */

start_server()
    {
    char *userargv[2];

    /* Create message queues for communication with server,
       do this before spawning server so as to avoid a race
       condition */

    if ((smsgid = msgget(INKEY,0777|IPC_CREAT)) < 0)
         ERR("msgget call failed for input queue");

    if ((rmsgid = msgget(OUTKEY,0777|IPC_CREAT)) < 0)
         ERR("msgget call failed for output queue");

    /* Start the lock manager process */

    user = 0;
    mypid = getpid();

    userargv[0] = "ioserver";
    userargv[1] = NULL;

    if ((serverpid = run("ioserver",userargv)) < 0)
         ERR("Could not start ioserver");

    if (server_request(INIT_SERVER_MSG)) ERR("ioserver init failed");

    printf("Started ioserver");
    }

stop_server()
    {
    register struct msgtime *tp;

    /* Extract and return CP and IO times */

    if (server_request(SERVER_TIMES_MSG)) server_usert = server_syst = 0.0;
    else {
         tp = (struct msgtime *)&rmsg;

         server_usert = tp -> usert;
         server_syst = tp -> syst;
         }

    /* Request ioserver to finish */

    if (server_request(TERM_SERVER_MSG))
         ERR("ioserver terminate failed");

    /* Close down the message queues */

    msgctl(smsgid,IPC_RMID,0);
    msgctl(rmsgid,IPC_RMID,0);
    }
#endif

/* Initialize working files */

initialize_files()
    {
    register int i,n;
    struct rec rec;

    rec.counter = 1;
    for(i = 0; i < control.reclen; i++) rec.junk[i] = 0;

    for(i = 0; i < control.n_files; i++) {
         opens[i] = closes[i] = uses[i] = reads[i] = writes[i] = 0;
         sum_time[i] = sumsq_time[i] = 0.0e0;
         worst_rec[i] = worst_user[i] = -1;
         worst_time[i] = -1.0e0;

         if ((unit[i] = open(control.path[i],O_RDWR)) < 0) {
              if ((unit[i] = creat(control.path[i],CREAT_MODE)) < 0)
                   ERRS("Could not open/create %s",control.path[i]);
              }

#if PRIMOS
         /* Set file for N readers and N writers */

         if (n = setrwlock(control.path[i]))
              ERRSS("Could not set RW lock for %s, code = %d",control.path[i],n);
#endif
         if (populate(i,&rec) < 0) {
              MSGS("Could not populate %s",control.path[i]);
#if UNIX
              if (errno = ENOSPC) {
                   for(;i>=0;i--) {
                        MSGS("Deleting %s",control.path[i]);
                        unlink(control.path[i]);
                        }
                   }
#endif
              ERR("Free up space and rerun");
              }
         }
    }

/* Create and initialize the results file, and, as long as we're looping,
   delete any left over runfiles */

create_results_file()
    {
    register int i;
    int rfid,n,z;
    char *rfp;
    struct results results;

    /* Calibrate CPU work for each transaction */

    if (control.param) {
         cptimes(&usert1,&syst1);
         work(control.param);
         cptimes(&usert2,&syst2);
         results.work_cp = usert2-usert1;
         }
    else results.work_cp = 0.0e0;

    OFLOOP fprintf(of[z],"CP load/%s = %.3f secs\n",TRANS,results.work_cp);

#if UNIX
#ifndef GTOD
    if (!brief) {
         OFLOOP fprintf(of[z],
"\nWarning: UNIX time() routine is being used to time %ss and disk\n\
reads, which means such times have a granularity of one second!  The default\n\
Makefile for IOBENCH tries to use the BSD gettimeofday routine, which\n\
provides much finer granularity.  You must have removed the -DGTOD from the\n\
Makefile or used -UGTOD in your Makefile wrapper.  If your system supports\n\
gettimeofday or has an equivalent, you should use it rather than time.  You\n\
can suppress this message by using the -b option.\n",TRANS);
         }
#endif
#endif

    unlink(RESULTS);
    if ((rfid = creat(RESULTS,CREAT_MODE)) < 0) ERRS("Could not create %s",RESULTS);

#if PRIMOS
    if (n = setrwlock(RESULTS))
         ERRSS("Could not set RW lock for %s, code = %d",RESULTS,n);
#endif

    /* Initialize the RESULTS contents so we can ensure that background
       processes actually wrote their results properly */

    results.counter = -1;
    results.start = results.finish = 0;
    results.usert = results.syst = 0.0e0;
    results.n_files = control.n_files;
    results.n_recs = control.n_recs;
    results.n_users = control.n_users;
    results.locks_per_file = control.locks_per_file;
    results.reclen = control.reclen;
    results.lock_sleep_time = control.lock_sleep_time;
    results.duration = control.duration;
    results.param = control.param;

    for (i = 0; i <= 25; i++) results.resp_dist[i]=0;

    if (write(rfid,&results,sizeof(results)) < 0)        /* for user 0 */
         ERRS("Could not write record 0 of %s",RESULTS);

    /* Each background process creates a RUNFILEnn of mode 0600 when it starts
       running and changes mode to 0666 as the last thing it does.  This
       lets us tell when all background processes have stopped.  If
       any of these files are left from a previous run, delete them */

    for(i = 1;i <= control.n_users; i++) {
         finished[i] = 1;
         results.userno = i;

         if (write(rfid,&results,sizeof(results)) < 0)   /* for user i */
              ERRSS("Could not write record %d of %s",i,RESULTS);

         if (fexists(rfp = runfile(i)))
              if (unlink(rfp) < 0) ERRS("Could not delete %s",rfp);

         userlog[i] = NULL;
         userlogsize[i] = 0;
         }

    if (close(rfid) < 0) ERRS("Could not close %s",RESULTS);
    }

/* Start up each user as a background process taking following
   arguments:

         1: the user number,
         2: the pathname of copy of control info (so it can find
            pathnames of files),
         3: the pathname of SYNCFILE
         4: the pathname of RUNFILEnn
         5: the pathname of RESULTS
         6: seed to use for iorandom()

   Each of the users will wait until sync file is created,
   at which time they will start the benchmark proper.

   Return number of errors encountered. */

start_user_processes()
    {
    register int i,z;
    int errs = 0,rfid;
    char *userargv[8],temp[16],seedstr[32];

    for(i = 1; i <= control.n_users && !errs; i++) {
         userargv[0] = "iouser";
         sprintf(temp,"%d",i);
         userargv[1] = temp;
         userargv[2] = CTLFILE;
         userargv[3] = SYNCFILE;
         userargv[4] = runfile(i);
         userargv[5] = RESULTS;
         sprintf(seedstr,"%d",seeds[i-1]);
         userargv[6] = seedstr;
         userargv[7] = NULL;

         if ((pid[i] = run("iouser",userargv)) > 0) {
              if (debug) printf("User %d is process %d\n",i,pid[i]);
              finished[i] = 0;
              spawned_users++;
              }
         else {
#ifdef USE_FEWER
              /* Since we've been busy spawning lots of processes, assume
                 that failure to create a process means there aren't any
                 more.  Instead of quitting, continue the run with the
                 processes we've created thus far. */

              if ((of[1] = fopen(outname,"a")) == NULL) {
                   MSGS("Could not reopen %s",outname);
                   errs++;
                   break;
                   }

              if ((rfid = open(RESULTS,O_RDWR)) < 0) {
                   MSGS("Could not reopen %s",RESULTS);
                   errs++;
                   break;
                   }

              OFLOOP {
                   fprintf(of[z],"Could not initiate process for user %d\n",i);
                   fprintf(of[z],"Continuing run with USERS = %d\n",i-1);
                   }

              control.n_users = results.n_users = i-1;

              if (lseek(rfid,0,SEEK_SET) < 0) ERRS("Could not rewind %s",RESULTS);

              if (write(rfid,&results,sizeof(results)) < 0)
                   ERRS("Could not rewrite record 0 of %s",RESULTS);

              if (fclose(of[1])) ERRS("Could not close %s",outname);
              if (close(rfid) < 0) ERRS("Could not close %s",RESULTS);
#else
              errs++;
              while(--i  > 0) killuser(i,pid);
#endif
              }
         }

    return(errs);
    }

/* Wait for all user processes to finish.  As each user process
   finishes, read its log file */

#if UNIX
wait_for_users_to_finish()
    {
    register int i,n;
    int errs,done,found,termpid,stat_loc,mode,size;
    char *rfp;

    done = errs = 0;
    while (!done) {

         /* Wait for wait call to return pid of a terminating iouser
            process.  Ignore return with negative value, which means
            a signal was received */

         while((termpid = wait(&stat_loc)) < 0);

         /* Check to see if this is a process we started */

         found = 0;
         for (i = 1; i <= control.n_users && !found; i++) {
              spawned_users--;

              if (termpid == pid[i]) {
                   found++;

                   if (finished[i]) {
                        MSGSS("Oops .. got another term notice for user %d, pid %d",
                         i,termpid);
                        errs++;
                        }
                   else {
                        finished[i] = 1;

                        if (stat_loc) {
                             MSGSS("User %d did not finish normally, exit status was %d",
                              i,stat_loc);
                             errs++;
                             }
                        else {
                             if (getinfo((rfp = runfile(i)),&mode,&size) < 0) {
                                  MSGS("Could not get info about %s",rfp);
                                  errs++;
                                  }

                             if (mode != 0666) {
                                  MSGSS("Mode of %s should be 0666 but is %o",rfp,mode);
                                  errs++;
                                  }
                             else read_transaction_log(rfp,i,size);
                             }
                        }
                   }
              }

         if (!found) {
              MSGS("Oops ... got term notice for pid %d",termpid);
              errs++;
              }

         /* We're done if all users have terminated */

         done = 1;
         for (i = 1; i <= control.n_users; i++)
              if (!finished[i]) done = 0;

         }

    if (errs) ERR("Run terminated because of errors noted above");
    }
#endif

#if PRIMOS

#define MINSLEEP 15

wait_for_users_to_finish()
    {
    register int loops,i,n;
    int mode,size;
    char *rfp;

    n = control.duration + control.end_delay;
    if (n < MINSLEEP) n = MINSLEEP;

    sleep(n);

    /* Wait for all background processes to finish */

    for(loops = 0;loops < MAXLOOPS; loops++) {
         printf("Checking termination ...\n");
         n = 0;
         for(i = 1; i <= control.n_users; i++) {
              if (!finished[i]) {
                   if (getinfo((rfp = runfile(i)),&mode,&size) < 0)
                        ERRS("Could not get info about %s",rfp);
                   if (mode != 0666) n++;
                   else {

                        /* Mode change means this iouser process is done */

                        finished[i] = 1;

                        /* Read all log entries in the file */

                        read_transaction_log(rfp,i,size);
                        }

                   }
              }

         if (n == 0) break;

         sleep(30);
         }

    /* Make sure every iouser process finished */

    errs = 0;
    for(i = 1; i <= control.n_users; i++)
         if (!finished[i]) {
              MSGS("User %d may still be running",i);
              killuser(i,pid);
              errs++;
              }
    }
#endif

/* Read the transaction log file produced by user i */

read_transaction_log(rfp,i,size)
char *rfp;
int i,size;
    {
    register int j;
    struct logrec *lp;

    if ((lp = (struct logrec *)malloc(size)) == NULL)
         ERR("Could not malloc space for log file");

    if ((j = open(rfp,O_RDONLY)) < 0)
         ERRS("Could not open %s",rfp);

    if (read(j,lp,size) != size)
         ERRS("Could not read %s",rfp);

    if (close(j) < 0) ERRS("Could not close %s",rfp);

    userlog[i] = lp;
    }

/* Output the statistics that user processes left in results file.
   Statistics for user K are in record K. */

output_user_statistics(rfid)
int rfid;
    {
    register int i,j,n;
    int k,z;
    struct results results;
    struct checkpoint *cp;

    for(j=0;j <= NUMBER_OF_CHECKPOINTS;j++) {
         cp = &checkpoint[j];
         cp->counter = cp->delta = 0;
         cp->usert = cp->syst = 0.0;
         }

    for (i= 0; i < 26; i++) cum_resp_dist[i]=0;

    total_avg_resp = 0.0;
    ntrans = total_nc = total_no = failed = maxet = 0;
    usert1 = 0.0;
    syst1 = 0.0;
    fprintf(of[1],
    "\n USER  %s USER TIME  SYS TIME   OPENS  CLOSES  ----START TIME----   DURATION\n",NUMTRAN);

    for (i = 1;i <= control.n_users; i++) {
         getrec(rfid,i,&results,sizeof(results));

         if (results.counter < 0) {
              fprintf(of[1],"%5d did not write results\n",i);
              failed++;
              continue;
              }

         if (results.counter == 0) {
              fprintf(of[1],"%5d did not do any %ss\n",i,TRANS);
              failed++;
              continue;
              }

         if (results.userno != i) {
              fprintf(of[1],"%5d contains data for user %d\n",i,results.userno);
              errs++;
              continue;
              }

         userlogsize[i] = results.counter;

         nc = no = 0;

         for(j = 0; j < control.n_files;j++) {
              opens[j] += results.opens[j];
              no += results.opens[j];
              reads[j] += results.freads[j];
              writes[j] += results.fwrites[j];

              closes[j] += results.closes[j];
              nc += results.closes[j];

              uses[j] += results.uses[j];
              sum_time[j] += results.sum_time[j];
              sumsq_time[j] += results.sumsq_time[j];

              if (results.worst_time[j] > worst_time[j]) {
                   worst_time[j] = results.worst_time[j];
                   worst_rec[j] = results.worst_rec[j];
                   worst_user[j] = i;
                   }
              }

         for(j=1;j <= NUMBER_OF_CHECKPOINTS;j++) {
              cp = &checkpoint[j];
              cp->counter += results.checkpoint[j].counter;
              cp->delta += results.checkpoint[j].delta;
              cp->usert += results.checkpoint[j].usert;
              cp->syst += results.checkpoint[j].syst;
              }

         et=results.finish-results.start;
         if (et > maxet) maxet = et;
         avg_resp= (double) et / (double) results.counter;
         fprintf(of[1],"%5d  %5d  %8.3f  %8.3f%8d%8d  %s  %8d\n",i,results.counter,
          results.usert,results.syst,no,nc,cvtime(results.start,1),et);
         ntrans += results.counter;
         usert1 += results.usert;
         syst1 += results.syst;
         total_no += no;
         total_nc += nc;
         total_avg_resp = total_avg_resp + avg_resp;

         for (k=0;k<=25; k++) cum_resp_dist[k] +=results.resp_dist[k];
         }
    }

/* Write checkpoint data */

output_checkpoints()
    {
    register int j,k;
    register struct checkpoint *cp;

    fprintf(of[1],"\nCheckpoint data follows\n\n\
CP%s   SECS   USER T   SYS T  Delta %s  Aggregate %s\n",NUMTRANS,TPS,TPS);
    for(j=1;j <= NUMBER_OF_CHECKPOINTS;j++) {
         cp = &checkpoint[j];
         cp->delta /= control.n_users;
         fprintf(of[1],"%2d %6d%7d%9.3f%8.3f",
          j,cp->counter,cp->delta,cp->usert,cp->syst);
         if (k = cp -> delta - checkpoint[j-1].delta)
              fprintf(of[1],"%11.3f",
               (double)(cp->counter-checkpoint[j-1].counter)/(double)k);
         else fprintf(of[1],"   infinite");
         fprintf(of[1],"%15.3f\n",(double)(cp->counter)/(double)(cp -> delta));
         }
    }

/* Output statistics on file usage */

output_file_statistics()
    {
    register int j,z;
    double samples,avg_time,variance_time,stddev_time;

    OFLOOP fprintf(of[z],"\nFILE   OPENS  CLOSES    USES  AVG TIME   STD DEV     MAX T  RECORD  USER\n");
    for(j = 0;j < control.n_files; j++) {

         samples = (double)uses[j];
         avg_time = (double)sum_time[j]/samples;

         variance_time = (double)sumsq_time[j]/samples -  avg_time * avg_time;

         stddev_time = sqrt(variance_time);

         OFLOOP {
              fprintf(of[z],"%4d%8d%8d  %6d  %8.3f  ",j, opens[j],
                closes[j],uses[j],avg_time);
              fprintf(of[z],"%8.3f  %8.3f  %6d  %4d\n", stddev_time,worst_time[j],
                worst_rec[j],worst_user[j]);
              }
         }
    }

/* Output totals */

output_totals()
    {
    register int z;

    mb = ((double)(ntrans * (control.n_reads + control.n_writes) *
     control.reclen)) / 1.0E6;

    OFLOOP {
         fprintf(of[z],"\n*** Actual   time = %d seconds\n",maxet);
         fprintf(of[z],"*** Flush    time = %d seconds\n",ft);
         fprintf(of[z],"*** Adjusted time = %d seconds\n",totaltime);

         fprintf(of[z],"\n*** Total %ss = %d\n",TRANS,ntrans);
         fprintf(of[z],"*** Total MB of data = %.3f\n\n",mb);

         fprintf(of[z],"*** User       CP seconds = %8.3f\n",usert1);
         fprintf(of[z],"*** System     CP seconds = %8.3f\n\n",syst1);

         fprintf(of[z],"*** Server (U) CP seconds = %8.3f\n",server_usert);
         fprintf(of[z],"*** Server (S) CP seconds = %8.3f\n\n",server_syst);

         fprintf(of[z],"*** Flush (U)  CP seconds = %8.3f\n",flush_usert);
         fprintf(of[z],"*** Flush (S)  CP seconds = %8.3f\n\n",flush_syst);
         }

    usert1 += (server_usert+flush_usert);
    syst1 += (server_syst+flush_syst);
    totalcp = usert1 + syst1;

    OFLOOP {
         fprintf(of[z],"*** Total User   CP seconds = %8.3f (%4.1f%%)\n",
          usert1,100.0*usert1/totalcp);
         fprintf(of[z],"*** Total System CP seconds = %8.3f (%4.1f%%)\n",
          syst1,100.0*syst1/totalcp);
         fprintf(of[z],"*** Total        CP seconds = %8.3f\n",usert1+syst1);
         }
    }

/* Show transaction statistics */

output_transaction_statistics(rfid)
int rfid;
    {
    register int i,k,z;
    double ut,st,tt,samples,avg_time,variance_time,stddev_time,max_time;
    struct results results;

    OFLOOP {
         fprintf(of[z],"\n*** %s statistics ***\n",CAPTRANS);
         fprintf(of[z],"\n USER  AVG TIME   STD DEV  MAX TIME\n");
         }

    max_time = 0.0;
    for(i = 1; i <= control.n_users;i++) {

         getrec(rfid,i,&results,sizeof(results));

         samples = (double)results.counter;
         avg_time = results.tran_sum_time/samples;
         variance_time = results.tran_sumsq_time/samples - avg_time * avg_time;
         stddev_time = sqrt(variance_time);

         if (results.tran_max_time > max_time) max_time = results.tran_max_time;

         OFLOOP fprintf(of[z],"%5d  %8.3f  %8.3f  %8.3f\n",i,avg_time,stddev_time,
          results.tran_max_time);
         }

    tps = (double) ntrans / (double) totaltime;
    total_avg_resp = total_avg_resp / (double) control.n_users;
    ut = 1000.0 * usert1 / (double) ntrans;
    st = 1000.0 * syst1 / (double) ntrans;
    tt = 1000.0 * totalcp / (double) ntrans;

    OFLOOP {
         fprintf(of[z],"\n*** %d %ss consisting of %d reads and %d writes\n",ntrans,
          TRANS,control.n_reads,control.n_writes);
#ifdef SPEC
         fprintf(of[z],"*** Cycles per Second = %7.3f\n",tps);
         fprintf(of[z],"*** CPS per user      = %7.3f\n",
          tps/(double)control.n_users);
#else
         fprintf(of[z],"*** Transactions per Second = %7.3f\n",tps);
         fprintf(of[z],"*** TPS per user            = %7.3f\n",
          tps/(double)control.n_users);
#endif
         fprintf(of[z],"*** Average response time = %.3f Seconds\n",total_avg_resp);
         fprintf(of[z],"*** Maximum response time = %.3f Seconds\n",max_time);
         fprintf(of[z],"*** Average User   CP milliseconds per %s = %8.3f (%4.1f%%)\n",
          TRANS,ut,100.0*ut/tt);
         fprintf(of[z],"*** Average System CP milliseconds per %s = %8.3f (%4.1f%%)\n",
          TRANS,st,100.0*st/tt);
         fprintf(of[z],"*** Average Total  CP milliseconds per %s = %8.3f\n",
          TRANS,tt);

         if (hits) fprintf(of[z],"*** Total number of lock hits = %d\n",hits);
	 }

    if (show_tps) {

         /* Show response time details */

         fprintf(of[1],"\n*** Response time distribution table\n\n");
         fprintf(of[1],"Seconds |");
         for (k=0; k<=8; k++)    fprintf (of[1]," %3d  |",k);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Count   |");
         for (k=0; k<=8; k++)    fprintf (of[1],"%5d |",cum_resp_dist[k]);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Percent |");
         for (k=0;k<=8; k++) fprintf(of[1],"%4.1f%% |",cum_resp_dist[k]/(double) ntrans *100.0);
         fprintf(of[1],"\n\nSeconds |");
         for (k=9; k<=17; k++)    fprintf (of[1]," %3d  |",k);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Count   |");
         for (k=9; k<=17; k++) fprintf (of[1],"%5d |",cum_resp_dist[k]);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Percent |");
         for (k=9;k<=17; k++) fprintf(of[1],"%4.1f%% |",
          cum_resp_dist[k]/(double) ntrans *100.0);
         fprintf(of[1],"\n\nSeconds |");
         for (k=18; k<=24; k++)    fprintf (of[1]," %3d  |",k);
         fprintf(of[1]," >24  |");
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+\n");
         fprintf(of[1],"Count   |");
         for (k=18; k<=25; k++) fprintf (of[1],"%5d |",cum_resp_dist[k]);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+\n");
         fprintf(of[1],"Percent |");
         for (k=18;k<=25; k++) fprintf(of[1],"%4.1f%% |",
          cum_resp_dist[k]/(double) ntrans *100.0);
         fprintf(of[1],"\n");
         }
    }


/* This routine is called to adjust results when some user processes
   didn't finish properly */

adjust_results()
    {
    register int z;
    int avg_n,completed;
    double avg_usert,avg_syst;

    OFLOOP fprintf(of[z],"\n*** %d users did not complete normally\n",failed);

    completed = control.n_users - failed;

    avg_n = ntrans / completed;
    avg_usert = usert1 / (double)completed;
    avg_syst = syst1 / (double)completed;

    ntrans += failed * avg_n;
    usert1 += failed * avg_usert;
    syst1 += failed * avg_syst;

    mb = ((double)(ntrans * 2 * control.reclen)) / 1.0E6;

    OFLOOP {
         fprintf(of[z],"\n*** Adjusted Total %ss = %d\n",TRANS,ntrans);

         fprintf(of[z],"*** Adjusted Total MB of data = %.3f\n",mb);
         fprintf(of[z],"*** Adjusted Total CP seconds = %.3f\n",usert1);
         fprintf(of[z],"*** Adjusted Total IO seconds = %.3f\n",syst1);
         }
    }

/* Write line in summary file - the line contains a series of fields
   separated by tabs

    Field         Contents

      1           Number of files (input)
      2           Number of users (input)
      3           Record length (input)
      4           Number of records (input)
      5           Number of reads (input)
      6           Number of writes (input)
      7           CPU work parameter (input)
      8           Number of transactions (input or output)
      9           Transactions per second (output)
     10           Actual time (output)
     11           Flush time (output)
     12           User CP seconds (output)
     13           System CP seconds (output)
     14           Total average response time (output)
     15           Number of failed users (output)
     16           Number of lock hits (output)
*/

write_summary_line()
    {
    register int old,z;

    if (iosummary) {
         old = fexists(iosummary);

         of[2] = fopen(iosummary,APPEND_MODE);

         if (of[2] == NULL) ERRS("Could not open %s",iosummary);
         if (!old) fprintf(of[2],"files\tusers\treclen\tnrecs\treads\twrites\t\
work\t%s\t%ss\tta\ttf\tuser\tsys\tavg. rt\tfailed\thits\n",TRANS,TPS);
         }

    OFLOOP {
         if (z < 2) fprintf(of[z],"\n===");

         fprintf(of[z],"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d",
          control.n_files,control.n_users,control.reclen,control.n_recs,
          control.n_reads,control.n_writes,control.param,ntrans);
         if (failed+errs == 0)
              fprintf(of[z],"\t%.3f\t%d\t%d\t%.3f\t%.3f\t%.3f\t%d\t%d\n",
               tps,maxet,ft,usert1,syst1,total_avg_resp,failed,hits);
         else fprintf(of[z],"\t*** INVALID RESULTS ***\n");
         }

    if (iosummary) fclose(of[2]);       /* ignore error, we're almost done */
    }

/* Each iouser process wrote a log identifying the records that it
   wrote - each log entry for each user contains the user number,
   the file number, the record number, the transaction counter,
   and the time at which the record was written.

   The following code processes each of the user logs from newest
   entry to oldest entry checking that the data in the log entry
   corresponds to what is in the file.

   If the data in the record matches the log entry, we continue
   checking with the next earliest log entry.  If the data in the
   record doesn't match and was written by the current user, we
   have a validation error.  If the data in the record doesn't match
   and belongs to another user, the log entry for the other user is
   checked.  If the record data matches the log entry for the other
   user, the record data is OK.  If the record data doesn't match
   the log entry for the other user, we have a validation error.

   Return number of errors detected */

validate()
    {
    register int i,j,k;
    int errs,limit,first,last,found,z;
    struct rec rec;
    struct logrec *log1,*lp1,*log2,*lp2;

    if (control.n_writes == 0) return(0);

    errs = 0;
    for (i = 1; i <= control.n_users; i++) {

         log1 = userlog[i];
         limit = full ? 0 : userlogsize[i]-1;

         for (j = userlogsize[i] - 1; j >= limit; j--) {

              lp1 = &log1[j];
              getrec(unit[lp1 -> l_file],lp1 -> l_record, &rec, control.reclen);

              if (rec.user == i
               && rec.counter == lp1 -> l_counter
               && rec.time == lp1 -> l_time) {

                   /* Data matches */

                   }
              else {


                   /*  Data mismatches, look for this transaction in
                       log of user referenced in the file record.  Since
                       there is a log record for each write in the transaction,
                       we've got to search a range of entries */

                   first = rec.counter * control.n_writes;
                   last = first + control.n_writes - 1;

                   if (debug) OFLOOP fprintf(of[z],
"debug: possible VE for user %d, file %d, record %d, counter %d, time %d\n\
debug: data record contains count %d, time %d, user %d\n\
debug: starting check, first = %d, last = %d\n",
                        i,lp1 -> l_file,lp1 -> l_record,rec.counter,rec.time,
                        rec.counter,rec.time,rec.user,first,last);

                   log2 = userlog[rec.user];

                   found = 0;

                   for(k = first; k <= last; k++) {
                        lp2 = &log2[k];

                        if (debug) OFLOOP fprintf(of[z],
"debug: log says file %d, record %d, counter %d, time %d\n",
                             lp2 -> l_file, lp2 -> l_record, lp2 -> l_counter,
                             lp2 -> l_time);

                        if (rec.counter == lp2 -> l_counter
                         && lp1 -> l_file == lp2 -> l_file
                         && lp1 -> l_record == lp2 -> l_record
                         && rec.time == lp2 -> l_time) {

                             /* Found the transaction that wrote this record */

                             if (debug) OFLOOP fprintf(of[z],"debug: found it!\n");

                             found++;
                             }
                        }

                   if (!found) {
                        validation_error(i,lp1,&rec);
                        had_validation_error++;
                        errs++;
                        }
                   }              /* if ... */

              }                   /* for(j... */

         }                        /* for(i... */

    return(errs);
    }

/* Print validation error message */

validation_error(user,lp,rp)
int user;
struct logrec *lp;       /* Pointer to log rec where error detected */
struct rec *rp;
    {
    register int z;

    OFLOOP {
         fprintf(of[z],"Validation failed for user %d, file %d, record %d\n\
Expected count %d, time %d\n\
   Found count %d, time %d, user %d\n",
          user,lp -> l_file,lp -> l_record,lp -> l_counter, lp -> l_time,
          rp -> counter, rp -> time, rp -> user);
         if (rp -> time < start_time) fprintf(of[z],"Record appears to have been written %d seconds before start of run!\n",
          start_time - rp -> time);
          }
    }

/* Delete files created during run that normally shouldn't be left behind */

delete_run_files()
    {
    register int i,z;
    char thedir[32];

    /* We're near the end of the run, so ignore errors */

    if (!keep)  {
         unlink(RESULTS);

         if (!had_validation_error)
              for(i=1; i<= control.n_users; i++) unlink(runfile(i));
         }

    unlink(CTLFILE);
    }

/* Return name of ith runfile */

char *runfile(i)
register int i;
    {
    static char rfname[32];

    sprintf(rfname,"%s%d",RUNFILE,i);
    return(rfname);
    }

/* Kill specified user */

killuser(un,pid)
int un,*pid;
    {
    if (kill(pid[un],SIGTERM) == 0) {
        if (debug)  fprintf(stderr,"Killed user %d, process %d\n",
          un,pid[un]);
        }
    else MSGSS("Could not kill user %d, process %d",un,pid[un]);
    }

#if PRIMOS
fortran
#endif
void cleanup(signo)
int signo;
    {
    register int i;

    if (inited_locks) termlocks();

    delete_run_files();

#if PRIMOS
    /* Kill user processes */

    if (spawned_users)
         for(i = 1; i <= control.n_users; i++)
              if (!finished[i])
                   killuser(i,pid);
#endif
    }

#if PRIMOS
/* Catch the PH_LOGO$ condition signalled when phantom logs out,
   this eliminates phantom logout messages on std output.  Global
   variable spawned_users is decremented so main can tell when all
   notices have been received */

fortran void ignore_logouts(infopt)
short *infopt;
    {
    short *p,msg[7],status,code;

    p = msg;
    do {
         lon$r(p,(short)6,status,code);
         spawned_users--;
         }
    while(status);
    }
#endif
