/* ioparams.c
/* Copyright (c) 1990, Prime Computer, Inc., Natick, MA 01760
/*                     All Rights Reserved
*/

/* Routines in this module process iobench parameters

   Initial Version: 27 June 1990 by Barry Wolman

          Modified: 16 July 1990 by Barry Wolman
               1. Add MOUNTS parameter
               2. Use special for FILEnn, DIRnn, DEVICEnn
               3. Init *_spec variables

          Modified: 24 July 1990 by Barry Wolman
               1. Use ERR macros instead of gripe in some cases

          Modified: 6 August 1990 by Barry Wolman
               1. Add error checks
               2. Use MSG macros instead of fprintf(stderr,...)

          Modified: 8 November 1990 by Barry Wolman
               1. Add some error checks

*/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "ioincl.h"
#include "iodefaults.h"

extern struct control control;

int duration_spec, trans_spec, nrecs_spec, filelen_spec;

/* Following table defines keywords recognized in control file. */

typedef struct {
    char *name;
    int  *valueptr;
    int  default_value;
    int  *specptr;
    } keyword_table;

keyword_table keyword[] = {
    {"RUN",        &control.runid,          DEFAULT_RUNID,       NULL},
    {"FILES",      &control.n_files,        DEFAULT_FILES,       NULL},
    {"USERS",      &control.n_users,        DEFAULT_USERS,       NULL},
    {"NRECS",      &control.n_recs,         DEFAULT_NRECS,       &nrecs_spec},
    {"FILELEN",    &control.filelen,        DEFAULT_FILELEN,     &filelen_spec},
    {"RECLEN",     &control.reclen,         DEFAULT_RECLEN,      NULL},
    {"LOCKS",      &control.locks_per_file, DEFAULT_LOCKS,       NULL},
    {"SLEEP",      &control.lock_sleep_time,DEFAULT_SLEEP,       NULL},
    {"START",      &control.start_delay,    DEFAULT_START,       NULL},
    {"END",        &control.end_delay,      DEFAULT_END,         NULL},
    {"DURATION",   &control.duration,       DEFAULT_DURATION,    &duration_spec},
    {"WORK",       &control.param,          DEFAULT_WORK,        NULL},
    {"READS",      &control.n_reads,        DEFAULT_READS,       NULL},
    {"WRITES",     &control.n_writes,       DEFAULT_WRITES,      NULL},
    {"TRANS",      &control.trans,          DEFAULT_TRANS,       &trans_spec},
    {"MOUNTS",     &control.mounts,         DEFAULT_MOUNTS,      NULL},
    NULL};

/* Following table defines parameters requiring special processing */

struct special {
    char     *prefix;
    char     **array;
    } special[] = {
    {"FILE",      control.path},
    {"DIR",       control.dir},
    {"DEVICE",    control.device},
    NULL};

char *runtype_str[] = { "DURATION", "TRANS"};

char *getenv(), *copystr();

/* Initialize parameters */

init_params()
    {
    register int i,j;
    register char *p;
    register keyword_table *kp;
    char temp[16];

    /* Initialize value from environment or from default value */

    nrecs_spec = filelen_spec = duration_spec = trans_spec = 0;

    for(kp = keyword; kp -> name; kp ++)
         if (p = getenv(kp -> name)) {
             *(kp -> valueptr) = atoi(p);
             if (kp -> specptr) *(kp -> specptr) = 1;
             }
         else *(kp -> valueptr) = kp -> default_value;

    for (i = 0; special[i].prefix; i++)
         for (j= 0; j < MAX_FILES; j++) {
              sprintf(temp,"%s%d",special[i].prefix,j);
              special[i].array[j] = (p = getenv(temp)) ? copystr(p) : NULL;
              }
    }

/* Parse control file with keyword assignments of the form
         FILES = 4; USERS = 64; ...  */

#define TEMPLEN 256

parse_control(cfname)
char *cfname;
    {
    register FILE *fid;
    register int c;
    register keyword_table *kp;
    int errs,n,found,i;
    char namestr[128],valuestr[128],temp[TEMPLEN],*np,*vp;

    if ((fid = fopen(cfname,"r")) == NULL)
         ERRS("Could not open control file %s",cfname);

    errs = 0;

    while(1) {

         /* Skip over optional white space */

         while(isspace(c = getc(fid)) && (c != EOF));

         if (c == EOF) break;

         if (c == '@') {

              /* Have switch to secondary file (which better not be empty) */

              if (fgets(temp,TEMPLEN,fid) == NULL) {
                   MSGS("Premature EOF in control file %s",cfname);
                   if (fclose(fid)) ERRS("Could not close %s",cfname);
                   return(1);
                   }

              temp[strlen(temp)-1] = 0;          /* Drop NL at end */

              if (fclose(fid)) ERRS("Could not close %s",cfname);

              if ((fid = fopen(temp,"r")) == NULL)
                   ERRS("Could not open indirect file %s",temp);

              while(isspace(c = getc(fid)));
              }

         /* Look for 'name=value;' with optional white space */

         np = namestr;
         while((c != '=') && !isspace(c)) {
              *np++ = islower(c) ? toupper(c) : c;
              c = getc(fid);
              }

         *np = 0;

         while(isspace(c)) c = getc(fid);

         if (c != '=') {
              MSG("'=' not found where expected");
              errs++;
              }

         while(isspace(c = getc(fid)));

         vp = valuestr;
         while((c != ';') && !isspace(c)) {
              *vp++ = c;
              c = getc(fid);
              }

         *vp = 0;

         while(isspace(c)) c = getc(fid);

         if (c != ';') {
              MSG("';' not found where expected");
              errs++;
              }

         /* First check to see if we have special parameter, e.g. FILEnn */

         found = 0;
         for(i = 0; !found && special[i].prefix; i++) {
              sprintf(temp,"%s%s",special[i].prefix,"%d");
              if (sscanf(namestr,temp,&n) == 1) {
                   if (n >= MAX_FILES) {
                        ERRSSS("Cannot set %s%d, max number of files is %d",
                         special[i].prefix,n,MAX_FILES);
                        errs++;
                        }
                   else special[i].array[n] = copystr(valuestr);
                   found = 1;
                   }
              }

         if (!found) {

              /* Check parameter name table */

              found = 0;
              for(kp = keyword;kp -> name && !found;)
                   if (strcmp(kp -> name,namestr)==0) found++;
                   else kp++;

              if (!found) {
                   MSGS("'%s' is not a valid parameter name",namestr);
                   errs++;
                   }
              else {
                   *(kp -> valueptr) = atoi(valuestr);
                   if (kp -> specptr) *(kp -> specptr) = 1;
                   }
              }
         }

    if (fclose(fid)) ERRS("Could not close %s",cfname);

    return(errs == 0 ? 0 : 1);
    }

/* Check validity of specified parameters */

check_params()
    {
    register int errs = 0,i;

    if (control.n_files <= 0) {
         value_error("Number of files",">",control.n_files);
         errs++;
         }

    if (control.n_files > MAX_FILES) {
         MSGSS("Maximum of %d files allowed, you requested %d",
          MAX_FILES,control.n_files);
         rebuild_message("MAX_FILES");
         errs++;
         }

    for (i = 0; i < control.n_files; i++)
         if (control.path[i] == NULL) {
              MSGS("FILE%d was not specified",i);
              errs++;
             }

    if (control.n_recs <= 0) {
         value_error("Number of records",">",control.n_recs);
         errs++;
         }

    if (control.reclen <= 0) {
         value_error("Record length",">",control.reclen);
         errs++;
         }

    if (control.reclen > MAX_RECLEN) {
         MSGSS("Maximum record length is %d, you requested %d\n",
          MAX_RECLEN,control.reclen);
         rebuild_message("MAX_RECLEN");
         errs++;
         }

    if (control.filelen <= 0) {
         value_error("File length", ">",control.filelen);
         errs++;
         }

    if (control.n_users> MAX_USERS) {
         MSGSS("Maximum number of users is %d, you requested %d\n",
          MAX_USERS,control.n_users);
         rebuild_message("MAX_USERS");
         errs++;
         }

    if (control.n_users <= 0) {
         value_error("Number of users",">",control.n_users);
         errs++;
         }

    if (control.locks_per_file < 0) {
         value_error("Locks per file",">=",control.locks_per_file);
         errs++;
         }

    if (control.start_delay < 0) {
         value_error("Start delay",">=",control.start_delay);
         errs++;
         }

    if (control.end_delay < 0) {
         value_error("End delay",">=",control.end_delay);
         errs++;
         }

    if (control.n_reads <= 0) {
         value_error("Number of reads",">",control.n_reads);
         errs++;
         }

    if (control.n_writes < 0) {
         value_error("Number of writes",">=",control.n_writes);
         errs++;
         }

    if (control.trans <= 0) {
         value_error("Number of transactions", ">",control.trans);
         errs++;
         }

    if (control.n_reads < control.n_writes) {
         MSGSS("Number of writes (%d%) must be <= Number of reads (%d)",
          control.n_writes,control.n_reads);
         errs++;
         }

    if (control.mounts < 0) {
         value_error("Number of mounts", ">=",control.mounts);
         errs++;
         }

    if (control.mounts > MAX_FILES) {
         MSGSS("Maximum of %d mounts allowed, you requested %d",
          MAX_FILES,control.mounts);
         rebuild_message("MAX_FILES");
         errs++;
         }

    for (i = 0; i < control.mounts; i++) {
         if (control.dir[i] == NULL) {
              MSGS("DIR%d was not specified",i);
              errs++;
              }

         if (control.device[i] == NULL) {
              MSGS("DEVICE%d was not specified",i);
              errs++;
              }
          }

    /* Figure out whether to do timed or counted run */

    switch(duration_spec + trans_spec) { 

         /* Both specified - do counted run using specified duration */

case 2:  control.runtype = COUNTED_RUN;
         break;

         /* Neither specified, do default */

case 0:  control.runtype = DEFAULT_RUNTYPE;
         break;
       
case 1:  /* Only one specified */

         control.runtype = duration_spec? TIMED_RUN : COUNTED_RUN;
         }

     /* If this is counted run, we need to estimate how long the
        run will take if user didn't specify */

     if (control.runtype == COUNTED_RUN && !duration_spec)
          control.duration = control.trans / NOMINAL_TPS;

     /* Figure out whether user specified number of records or filelen */

     switch(nrecs_spec + filelen_spec) {

         /* Both specified, ignore FILELEN specification */

case 2:  MSG("Both NRECS and FILELEN specified, ignoring FILELEN");

         /* Neither specified */

case 0:  control.filelen = control.n_recs * control.reclen;
         break;

         /* Only one specified */

case 1:  if (nrecs_spec) control.filelen = control.n_recs * control.reclen;
         else {
              control.n_recs = control.filelen / control.reclen;
              if (control.n_recs == 0) {
                   MSGSS("Specified record length %d is greater than specified file length %d",
                    control.reclen,control.filelen);
                   errs++;
                   }
              }
         }

    return(errs);
    }

/* Print message about what value a parameter must have */

value_error(id,sign,val)
char *id,*sign;
int val;
    {
    MSGSSS("%s must be %s 0, you specified %d",id,sign,val);
    }

/* Print message about how to increase a benchmark limit */

rebuild_message(limit)
char *limit;
    {
    MSGS("To increase limit, change %s in 'ioincl.h' and remake IOBENCH",
     limit);
    }
