/* unix_subs.c
/* Copyright (c) 1988, Prime Computer, Inc., Natick, MA 01760
/*                     All Rights Reserved
*/

/* Note: if you get a syntax error on the line containing "clock_t",
   this means your system lacks a definition for the POSIX clock_t type.
   In this case, add "-Dclock_t=long" to EXTRA_CFLAGS in your Makefile
   wrapper and re-compile.

   UNIX versions of subroutines called by IO benchmark

         Modified:  4 April 1989 by Barry Wolman
              1. Add append flag to monitor_system

         Modified: 22 April 1990 by Barry Wolman
              1. Use execvp so that iouser and similar parts of
                 benchmark can be "installed".
              2. Change monitor program to call "iousage" instead
                 of "usage" and to do so in a way that "iousage"
                 can be "installed".

         Modified: 18 May 1990 by Barry Wolman
              1. Reorder args to monitor_usage and add delay parameter

         Modified: 20 June 1990 by Barry Wolman
              1. Rename systimes to cptimes, keep user & sys times
                 separate

         Modified: 16 July 1190 by Barry Wolman
              1. Add fullcptimes

         Modified: 18 July 1990 by Barry Wolman
              1. Change shm_ptr into sp (needed for IBM port)
              2. Handle case where HZ is not defined

         Modified: 23 July 1990 by Barry Wolman
              1. Remove #ifdef in declaration of times()

         Modified: 27 July 1990 by Barry Wolman
              1. Add getinfo

         Modified: 6 August 1990 by Barry Wolman
              1. Use ERR macros
              2. Add some error checking

         Modified: 10 August 1990 by Barry Wolman
              1. Fix use of shmat() to satisfy ANSI compilers

         Modified: 27 September 1990 by Barry Wolman
              1. Fix run to look in current directory if execvp call
                 fails (deals with case of root not having "." in
                 search rules).

         Modified: 8 October 1990 by Barry Wolman
              1. Handle case where HZ is defined as a sysconf call
*/

#include <stdio.h>
#include "ioincl.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>

/* Allocate 128KB of shared memory for use by the various users
   and return a pointer to it
*/

#define SH_MEM_SIZE     128*1024       /* Size of shared memory segment */
#define SH_MEM_KEY      4280           /* Arbitrary key for shared mem */

int shmid;                      /* Shared memory id and ptr */
char *sp;
extern char *sys_errlist[];

char *get_shared_memory(user)
int user;
    {
    key_t key = SH_MEM_KEY;

    if (user <= 0)   /* the control user or ioserver must create the sh mem segment */
        if (( shmid = shmget(key, SH_MEM_SIZE, (IPC_CREAT | IPC_EXCL | 00666))) < 0)
            ERR("Shared memory allocation failed");

    if (( shmid = shmget(key, SH_MEM_SIZE, 0)) < 0)
        ERRS("Shared memory allocation failed for user %d",user);

    if ((sp = (char *)shmat(shmid, (char *)0, 0)) == (char *)-1)
        ERRS("Shared memory attach failed for user %d",user);

    return(sp);
    }

/* Free shared memory segment.  Users 1-N just detach, removal is finally
   done when user 0 calls us */

int free_shared_memory(user)
int user;
    {
    if (user > 0) {

         /* This is not controlling user or ioserver */

         if (shmdt (sp) < 0)
             ERRS("Shared memory detach failed for user %d",user);
         }
    else {

         /* This is controlling user or ioserver */

         if (shmctl(shmid, IPC_RMID, NULL) < 0)
             ERR("Shared memory delete failed");
           }
    }

/*
   Run specified cmd with args as background process
   Last element of argv must be a NULL
*/

run(cmd,argv)
char *cmd, *argv[];
    {
    register int pid;
    char path[64];         /* More than enough */

    if (!(pid = fork()))
        if (execvp(cmd,argv) < 0) {
             if (errno != ENOENT) ERRS("exec failed for cmd %s",cmd);
             sprintf(path,"./%s",cmd);
             if (execv(path,argv) < 0) ERRS("exec failed for cmd %s",cmd);
             }

    /* Parent doesn't wait - we want child to run in background */

    return(pid);
    }

/* Generate background process to monitor system for specified
   duration putting output into specified path
*/

monitor_system(delay,duration,path,append)
int delay;       /* Seconds to wait before starting */
int duration;    /* Seconds to monitor */
char *path;      /* Where to put results */
int append;      /* Append flag */
    {
    char *monargv[7];
    char temp1[16],temp2[16];

    sprintf(temp1,"%d",delay);
    sprintf(temp2,"%d",duration);
    monargv[0] = "iousage";
    monargv[1] = temp1;
    monargv[2] = temp2;
    monargv[3] = path;
    monargv[4] = append ? "keep" : "rm";
    monargv[5] = NULL;

    run("iousage",monargv);
    }

/* Determine user and system CP times w/o counting child times */

#include <sys/times.h>
#include <sys/param.h>

/* If your system gives an error for the following definition, use
	-Dclock_t=long
   when compiling */

clock_t times();

/* Note: some systems, e.g. Motorola, define HZ as a call to sysconf(...),
   which prevents us from using
		int syshz = HZ;
*/

static int syshz = 0;

#ifdef HZ
#define SYSHZ HZ
#else
#include <unistd.h>
#define SYSHZ (getsyshz())

/* Determine number of clock ticks per second by using POSIX sysconf */

getsyshz()
    {
    if ((syshz = sysconf(_SC_CLK_TCK)) < 0) {
         MSG("Could not determine clock rate from HZ or sysconf.\n\
Using 100 - user and system CP times are likely to be meaningless.\n\
Wall clock times and TPS figures still will be correct.");
         syshz = 100;
         }
    }
#endif

cptimes(userp,sysp)
double *userp;    /* where to put user CP time */
double *sysp;     /* where to put system CP time */
    {
    struct tms tm;

    time_t tmp;

    if (syshz == 0) syshz = SYSHZ;

    times(&tm);

    *userp = ((double) tm.tms_utime)/syshz;
    *sysp  = ((double) tm.tms_stime)/syshz;
    }

/* Determine user and system CP times counting child procs */

fullcptimes(userp,sysp)
double *userp;   /* where to put user CP time */
double *sysp;    /* where to put system CP time */
    {
    struct tms tm;

    time_t tmp;

    if (syshz == 0) syshz = SYSHZ;

    times(&tm);

    *userp = ((double) (tm.tms_utime + tm.tms_cutime)) / syshz;
    *sysp  = ((double) (tm.tms_stime + tm.tms_cstime)) / syshz;
    }

/* Return size of file in bytes */

fsize(path)
char *path;
{
    struct stat sbuf;

    if ((stat(path,&sbuf)) < 0)
        ERRS("Could not stat file %s",path);

    return(sbuf.st_size);
}

/* Get mode (lower 9 bits) and size of file in bytes */

getinfo(path,mp,sp)
char *path;
int *mp,*sp;
    {
    struct stat sbuf;

    if ((stat(path,&sbuf)) < 0) return(-1);

    *mp = sbuf.st_mode & 0777;
    *sp = sbuf.st_size;
    return(0);
    }

